!Do Metropolis-Hastings and slice sampling algorithms
!Also directional gridding method for fast-slow parameters
!Provisional implementation of Wang-Landau and Multicanonical methods
!August 2006

module MonteCarlo
 use ParamDef
 use CalcLike
 use Random
 use propose
use cosmonet

 ! Nested sampling includes
 use nested_cluster
 use rand
 use sampling
 use tree
 use ellipsoid




 implicit none
 logical time, first_time
 type(ParamSet)  Trial
 logical :: calc_evidence_post = .True.
 real :: ExtremeLike
  !number of iterations between dumping full model data. If zero then skip.
                                                                                                   
#define rk 8
                                                                                                   
real(rk), dimension(num_params) :: p_min
real(rk), dimension(num_params) :: p_max
                                                                                                   
 REAL(rk) :: min_inf = -1.0 * HUGE(real(1.0, rk))




 integer :: indep_sample = 0
  !number of iterations between dumping full model data. If zero then skip.


 integer :: directional_grid_steps = 20 
  !for sampling_method = sampling_slowgrid, number of steps per grid

 real, parameter    :: eps_1 = 1.00001
 real :: MaxLike


!Multicanonical/W-L sampling parameters
  real, parameter :: mc_logspace =  6. !1/ steps to use in log-likelihood * number of parameter
  integer :: mc_mini = 0, mc_maxi = 0
  integer :: mc_update_steps = 7000, mc_update_burn = 50
  integer :: mc_steps_inc = 1000
  real, dimension(:,:), allocatable :: mc_like_counts
  real, dimension(:), allocatable :: mc_lnweights
!Wang-Landau parameter
  real :: WL_f = 1, WL_min = 0.1, WL_minW
  real :: WL_maxL=8. !WL_maxL is log like from best to use, in units of number of parameters
  integer :: WL_update_steps = 4000
  real :: WL_flat_tol = 1.4 !factor by which smallest histogram can be smaller than mean
contains

 function UpdateParamsLike(Params, fast, dist, i, freeNew, Lik) result(NewLik) 
  !For slice sampling movement and likelihood
  Type(ParamSet) tmp, Params
  integer, intent(in) :: i
  real, intent(in), optional :: Lik
  real NewLik
  logical, intent(in) :: fast, freeNew
  real, intent(in) :: dist

  tmp = Params

  call UpdateParamsDirection(tmp, fast, dist, i)

  NewLik = GetLogLike(tmp)
  if (freeNew) then
    call AcceptReject(.true., tmp%Info, Params%Info)
  else
     if (NewLik <= Lik*eps_1) then 
       !accept move
       call AcceptReject(.false., tmp%Info, Params%Info)
       Params = tmp
     else
       call AcceptReject(.true., tmp%Info, Params%Info)
     endif
  endif

 end function UpdateParamsLike


 subroutine SliceUpdate(Params, fast, i, Like)
 !Do slice sampling, see http://www.cs.toronto.edu/~radford/slice-aos.abstract.html
 !update parameter i, input Like updated to new value.
 !Use linear stepping for the moment
  Type(ParamSet) Params
  integer, intent(in) :: i
  logical, intent(in) :: fast
  real, intent(inout) :: Like
  real offset,  P
  real L, R 
  real LikL, LikR, Range, LikT, w
  integer fevals


   w = propose_scale
   if (slice_randomsize) w =w * Gaussian1()
   Like = Like + randexp1()  !New vertical position (likelihood)  
   fevals = 0

   offset = ranmar() 
 
   L = -offset
   R = 1- offset

   !step out
   if (slice_stepout) then
       do
        LikL = UpdateParamsLike(Params, fast, w*L, i, .true.)
        fevals = fevals + 1
        if (LikL*eps_1 > Like) exit
        L = L  - 1
       end do

       do
        LikR = UpdateParamsLike(Params, fast, w*R, i,.true.)
        fevals = fevals + 1
        if (LikR*eps_1 > Like) exit
        R = R + 1
       end do
   end if
     
   !stepping in
   do
    Range =  R - L
    P  =  ranmar() * Range + L
    LikT = UpdateParamsLike(Params, fast, w*P, i, .false., Like)
    fevals = fevals + 1
    if (LikT < Like*eps_1) exit
    if (P > 0) then
        R = P
    else
        L = P
    end if
   end do
   Like = LikT
   if (.not. fast) slow_proposals = slow_proposals + fevals
 
 end subroutine SliceUpdate


 subroutine SliceSampleSlowParam(CurParams, CurLike) 
  Type(ParamSet) CurParams
  real CurLike
  integer, save:: loopix = 0

   if (mod(loopix,num_slow)==0) then
        if (.not. allocated(Rot_slow)) allocate(Rot_slow(num_slow,num_slow))
        call RotMatrix(Rot_slow,num_slow)
        loopix = 0
   end if
   loopix = loopix + 1
   call SliceUpdate(CurParams, .false., loopix,CurLike) 

 end subroutine SliceSampleSlowParam
 
 subroutine SliceSampleFastParams(CurParams, CurLike)
  Type(ParamSet) CurParams
  real CurLike
  integer j 

   if (.not. allocated(Rot_fast)) allocate(Rot_fast(num_fast,num_fast))
   call RotMatrix(Rot_fast,num_fast)
   
   do j = 1, num_fast * max(1,oversample_fast)
      call SliceUpdate(CurParams, .true., mod(j-1,num_fast)+1,CurLike) 
   end do
   
 end subroutine SliceSampleFastParams



 subroutine SampleSlowGrid(CurParams, CurLike, mult, MaxLike, num, num_accept) 
  !sampling_method = method_slowgrid
  !Make proposals on a grid in each slow direction simultaneously with changes to
  !fast parameters. Allows fast paraters to 'adapt' to slow parameter values, thereby
  !increasing slow acceptance rate
  Type(ParamSet) Trial, CurParams
  real CurLike, MaxLike
  integer num, num_accept, mult
  real w
  integer r, r_min, r_max, last_r
  Type(ParamSet), dimension(:), allocatable :: grid
  integer i, accpt
  real Like
  integer, save:: loopix = 0


   if (mod(loopix,num_slow)==0) then
        if (.not. allocated(Rot_slow)) allocate(Rot_slow(num_slow,num_slow))
        call RotMatrix(Rot_slow,num_slow)
        loopix = 0
   end if
   loopix = loopix + 1

   w = max(1e-5,propose_r(num_slow)) * propose_scale / 2
 
   r = 0
   r_min = 0
   r_max = 0
   accpt = 0

   allocate(grid(-directional_grid_steps:directional_grid_steps))
   grid(0) = CurParams
   
   do i = 1, directional_grid_steps*oversample_fast

    last_r = r

    if (ranmar() < 0.5) then
     r =r  + 1
    else
     r =r  - 1
    end if
  
    Trial = Grid(last_r)
    call UpdateParamsDirection(Trial, .false., (r-last_r)*w, loopix)
 
    if (r > r_max  .or. r < r_min) then
     grid(r) = Trial
     slow_proposals = slow_proposals + 1
     r_min = min(r,r_min)
     r_max = max(r,r_max)
    else
     grid(r)%P(fast_params_used) = Trial%P(fast_params_used)
    end if

    call GetProposalProjFast(grid(r), grid(r), propose_scale)
  
    Like = GetLogLike(grid(r))
   
    if (Feedback > 1) write (*,*) r, 'Likelihood: ', Like, 'Current Like:', CurLike

    if ((Like /= logZero) .and. (CurLike > Like .or. randexp1() > Like - CurLike)) then
    !Accept
 
      if (num_accept> burn_in) then
          output_lines = output_lines +1
         call WriteParams(grid(last_r),real(mult),CurLike)
      end if
   
      CurLike = Like
      accpt = accpt + 1
      if (Like < MaxLike) MaxLike = Like
      mult=1

    else
    !Reject
      r = last_r
      mult=mult+1
    end if  

    call AddMPIParams(grid(r)%P,CurLike) 
 
   num = num + 1

   end do
   
   if (Feedback > 1) write(*,*) 'Done slow grid, direction', loopix
   if (Feedback > 1) write(*,*) 'grid steps moved:',abs(r), ' acc rate =',&
            real(accpt)/(directional_grid_steps*oversample_fast)
   
   !Have to be careful freeing up here.. (logZero results don't allocate new)
   do last_r = r_max, r+1, -1
      call AcceptReject(.true., grid(last_r)%Info,grid(last_r-1)%Info)  
   end do
   do last_r = r_min, r-1
      call AcceptReject(.true., grid(last_r)%Info,grid(last_r+1)%Info)  
   end do

   CurParams = grid(r)
   num_accept = num_accept + accpt

   deallocate(grid)

 end subroutine SampleSlowGrid

!Multicanonical/Wang-Landau

 subroutine MC_AddLike(L)
  real, intent(in) :: L
  integer like_ix
  integer, dimension(:,:), allocatable :: tmp
  integer sc

   sc =0

   like_ix = int(L*mc_logspace/num_params_used)
   if (.not. allocated(mc_like_counts)) then
     mc_mini=like_ix-100
     mc_maxi=like_ix+100
     allocate(mc_like_counts(mc_mini:mc_maxi,2))
     mc_like_counts=sc
   end if
   if (like_ix > mc_maxi) then
     allocate(tmp(mc_mini:like_ix+100,2))
     tmp=sc
     tmp(mc_mini:mc_maxi,:) = mc_like_counts
     deallocate(mc_like_counts)
     mc_maxi = like_ix + 100
     allocate(mc_like_counts(mc_mini:mc_maxi,2))
     mc_like_counts = tmp
     deallocate(tmp)
   end if
   if (like_ix < mc_mini) then
     allocate(tmp(mc_mini-100:mc_maxi,2))
     tmp=sc
     tmp(mc_mini:mc_maxi,:) = mc_like_counts
     deallocate(mc_like_counts)
     mc_mini = like_ix - 100
     allocate(mc_like_counts(mc_mini:mc_maxi,2))
     mc_like_counts = tmp
     deallocate(tmp)
   end if

   mc_like_counts(like_ix,1)=mc_like_counts(like_ix,1)+ WL_f
   mc_like_counts(like_ix,2)=mc_like_counts(like_ix,2)+ 1


 end subroutine MC_AddLike

 subroutine MC_UpdateWeights
    real, dimension(:), allocatable :: tmp
    real win(-2:2)
    integer i
         
    allocate(tmp(mc_mini:mc_maxi))
    tmp=0
    tmp(LBOUND(mc_lnweights,DIM=1):UBOUND(mc_lnweights,DIM=1)) = mc_lnweights   
    if (allocated(mc_lnweights)) deallocate(mc_lnweights)
    allocate(mc_lnweights(mc_mini:mc_maxi))
    mc_lnweights = tmp

    tmp = 0
    do i=-2, 2
     win(i) = exp(-real(i**2)*2)
    end do
    do i = mc_mini+2 , mc_maxi-2
       tmp(i) = sum(mc_like_counts(i-2:i+2,1)*win)
    end do 
    !tmp=mc_like_counts
    tmp = max(tmp,1.)
    mc_lnweights = mc_lnweights + log(tmp/real(maxval(tmp))) 
    deallocate(tmp)
    mc_like_counts = 0

 end  subroutine MC_UpdateWeights
 
 function MC_LnWeight(L)
  real, intent(in) :: L
  real MC_LnWeight
  integer like_ix

  if (.not. allocated(mc_lnweights)) then
       MC_LnWeight=0
  else
      like_ix = int(L*mc_logspace/num_params_used)
      if (like_ix > UBound(mc_lnweights,dim=1)) then
       MC_LnWeight = mc_lnweights(UBound(mc_lnweights,dim=1))    
      elseif (like_ix < LBound(mc_lnweights,dim=1)) then
       MC_LnWeight= mc_lnweights(LBound(mc_lnweights,dim=1))    
      else
       MC_LnWeight = mc_lnweights(like_ix)    
      end if
  endif

 end  function MC_LnWeight


 function MC_WeightLike(L) result (Like)
  real, intent(in) :: L
  real Like

   Like = L + MC_LnWeight(L)

 end function MC_WeightLike


 function MC_Weight(L) result(W)
  real, intent(in) :: L
  real mx, W
  integer like_ix

  if (.not. allocated(mc_lnweights)) then
   W=1
  else
   like_ix =int(L*mc_logspace/num_params_used)

   mx = maxval(mc_lnweights)
   W =  exp((MC_LnWeight(L) - mx))
  end if

 end function MC_Weight


 subroutine WL_UpdateWeights
  integer i,j,k, n
  real amin,asum,L

   if (WL_F > WL_min) then

     amin = LogZero
     asum = 0
     n=0
     do i=mc_mini,mc_maxi
      L = i*num_params_used/mc_logspace 
      if (mc_like_counts(i,1) /= 0 .and. L <= MaxLike + WL_maxL*num_params_used) then
           n=n+1
           amin = min(amin,mc_like_counts(i,2))
           asum = asum + mc_like_counts(i,2)
      end if
     end do
    WL_update_steps = WL_update_steps + 30*num_params_used 

    if (amin > asum / (WL_flat_tol*n)) then
          WL_f = WL_f*0.5
          mc_like_counts(:,2)=0
          WL_update_steps = WL_update_steps + 300*num_params_used  

          if (WL_F < WL_min) then
             if (Feedback > 0) write(*,*) 'Outputting Markovian samples'
             WL_f = 0
             WL_minW=LogZero
!            call CreateTxtFile('density.txt',1)
             do i=mc_mini,mc_maxi
              L = i*num_params_used/mc_logspace 
              if (mc_like_counts(i,1) /= 0 .and. L <= MaxLike + WL_maxL*num_params_used) then
                     WL_minW = min(WL_minW,L - mc_like_counts(i,1))
  !                   write (1,*) L,mc_like_counts(i,1)
              end if
             end do
 !            close(1)
          end if
    end if
   end if
 end  subroutine WL_UpdateWeights

 function WL_WeightLike(L) result (Like)
  real, intent(in) :: L
  real Like
  integer like_ix

  if (L > MaxLike + WL_maxL*num_params_used) then
    Like = LogZero  !!!!
    return
  end if

  like_ix = int(L*mc_logspace/num_params_used)
  if (like_ix >= mc_maxi .or. like_ix <= mc_mini) then
   Like = 0     
  else
   Like = mc_like_counts(like_ix,1)    
  end if

 end function WL_WeightLike


function WL_Weight(L) result (W)
  real, intent(in) :: L
  real W
  integer like_ix

  if (L > MaxLike + WL_maxL*num_params_used) then
   W=0
   return
  end if

  like_ix = int(L*mc_logspace/num_params_used)
  if (like_ix >= mc_maxi .or. like_ix <= mc_mini) then
   W = 0     
  else
   W  = exp( -L + mc_like_counts(like_ix,1) + WL_minW)    
  end if

 end function WL_Weight


 subroutine MCMCsample(Params, samples_to_get)
   integer samples_to_get
   Type(ParamSet) Trial, Params, CurParams
   real Like, CurLike
   logical accpt, time, first_time, within_range
   real rmult
   integer mult,num_accept, last_num
   integer num_metropolis,  num_metropolis_accept
   real testlike, testCurLike
   integer mc_update_burn_num, mc_updates, i
   real :: net_params(1:2*num_params_used)

 character(LEN =30) fmt
time = .true.
first_time = .true.
   MaxLike = LogZero
   CurLike = StartLike
   testCurLike = StartLike
   CurParams = Params
   num_accept = 0
   last_num = 0
   mult= 1
   num_metropolis = 0
   num_metropolis_accept = 0
   mc_update_burn_num = mc_update_burn
   mc_updates = 0

!!!!!!!!!!!!!!!!!!!!!!10param 4 sigma region!!!!!!!!!!!!
net_params(1) = 0.0205637d0
net_params(2) = 0.0801418d0
net_params(3) = 1.03216d0
net_params(4) = 0.0100661d0
net_params(5) = -0.145745d0
net_params(6) = 0.000000d0
net_params(7) = -1.50000d0
net_params(8) = 0.90911d0
net_params(9) = 2.81763d0
net_params(10) = 0.000000d0
net_params(11)=0.0262164d0
net_params(12)=0.141981d0
net_params(13)=1.05557d0
net_params(14)=0.181053d0
net_params(15)= 0.0719305d0
net_params(16) = 0.136317d0
net_params(17) = -0.0955112d0
net_params(18)=1.07102d0
net_params(19)=3.22724d0
net_params(20) = 0.606624d0


		FScales%PMin = Scales%PMin
		FScales%PMax = Scales%PMax
                !do i=1, num_params_used
                !        FScales%PMin(params_used(i)) = net_params(i)
                !        FScales%PMax(params_used(i)) = net_params(i+num_params_used)
                !        write(*,*) "FSCALES MCMC", FScales%PMin(params_used(i)), FScales%PMax(params_used(i))
                !end do

!FScales%PMin(1) = 0.0220237d0
!FScales%PMin(2) = 0.102716d0
!FScales%PMin(3) = 1.03979d0
!FScales%PMin(4) = 0.0702301d0 
!FScales%PMin(5) = -0.0540575d0
!FScales%PMin(6) = 0
!FScales%PMin(7) = -1 -1.50000d0 
!FScales%PMin(8) = 0.947561d0
!FScales%PMin(9) = 0
!FScales%PMin(10) = 0
!FScales%PMin(11) = 2.95868d0
!FScales%PMin(12) = 0
!FScales%PMin(13) = 0
                                                                           
                                                                           
!FScales%PMin(14) = 0.0237929d0
!FScales%PMin(15) = 0.119064d0
!FScales%PMin(16) = 1.04660d0
!FScales%PMin(17) = 0.0962470d0
!FScales%PMin(18) = 0.00461532d0 
!FScales%PMin(19) =  0.0418562d0
!FScales%PMin(20) = -0.636722d0
!FScales%PMin(21) = 0.996576d0
!FScales%PMin(22) = 0
!FScales%PMin(23) = 0
!FScales%PMin(24) = 3.07945d0
!FScales%PMin(25) = 0.118389d0
!FScales%PMin(26) = 0





!if (any(Trial%P > FScales%PMax) .or. any(Trial%P < FScales%PMin)) then
!	within_range = .false.
!	write(*,*) "Not using 1 sigma likes"
!else
	within_range = .true.
!	write(*,*) "Using 1 sigma likes"
!end if

open(14, FILE = "cosmonet_w_grid.dat")

do i=1, 150
        Trial%P(1) = 0.0246
        Trial%P(2) = 0.105
        Trial%P(3) = 1.04
        Trial%P(4) = 0.07
        Trial%P(5) = 0
        Trial%P(6) = 0
        Trial%P(7) = -1*i*0.01
        Trial%P(8) = 0.95
        Trial%P(11) = 3
        Trial%P(12) = 0.1
        fmt = trim(numcat('(2E16.7,',num_params))//'E16.7,7E16.7)'
        write(14,fmt) COSMONET_GetLogLike(Trial, time, first_time), Trial%P
end do




 end subroutine MCMCsample

subroutine Cluster_Nest_sample(Params, accept_num, rootname, init_widths)
        real :: init_widths(num_params, 2)
   real Like
   integer accept_num, i
   Type(ParamSet) Params, CurParams
   real CurLike, MaxLike
   logical accpt
   integer num_accept, num, mult, last_num
   character(LEN=80) rootname
   double precision ranmar
                                                                                                   
   Trial = Params
  do i=1, num_params
         p_min(i) = init_widths(i,1)
         p_max(i) = init_widths(i,2)
         write(*,*) "P_MIN",   init_widths(i,1), "P_MAX", init_widths(i,2)
 end do
                                                                                                   
call Nested_Evidence(Params, rootname)
                                                                                                   
 end subroutine Cluster_Nest_sample
                                                                                                   
subroutine Nested_Evidence(Params, rootname)
use sampling
   implicit none
   Type(ParamSet) Params
character(LEN =30) fmt
                                                                                                   
integer i,j,q
   real ParamsLike, final_weight, sum_of_weights, av
character(LEN=80) rootname
   double precision ranmar
                                                                                         
   REAL(rk), DIMENSION(1,1) :: data
   REAL(rk), DIMENSION(2) :: ev
   TYPE(pset), POINTER :: cs
   REAL(rk), DIMENSION(:,:), allocatable :: param
   REAL(rk), DIMENSION(6,500) :: param_test
                                                                                         
   REAL(rk), DIMENSION(10000) :: file_like
   REAL(rk), DIMENSION(10000) :: file_weight
   REAL(rk), DIMENSION(10000) :: x1
   REAL(rk), DIMENSION(10000) :: x2
   REAL(rk), DIMENSION(13) :: parameter
REAL(rk), allocatable, DIMENSION(:) :: wt
                                                                                                   
   REAL(rk) :: en
!   en = 1.0
en = 1.0
                                                                                                   
   print *, 'Starting Evidence calculation.'
                                                                                                   
   ! Set up sampling.
   call set_npoints(200)
!       call set_npoints(20)
   CALL set_nrepl(10000)
   CALL set_niter(10)
   CALL set_nevid(30)
!   CALL set_volfrac(real(1e-2,rk))
   CALL set_volfrac(real(0.1,rk))
   call set_volred(real(0.3,rk))
!   CALL set_clustsep(real(0.3,rk))
call set_clustsep(real(1.1,rk))
   CALL set_cluster(0)
   CALL set_ellskip(1)
   CALL set_cluster_output(.TRUE.)
   OPEN(74, FILE=trim(rootname)//"_cont_out.dat")
   CALL set_cont_output(74)
!open(88, FILE = trim(rootname)//"_resume.dat")
!call set_resume_fh(88)
 time = .true.
first_time = .true.
 
   CALL init_ellipsoid(data, 1, num_params_used, en, Nested_Like, Nested_Prior)
                                                                                                   
   ! Sample
   CALL sample_points(data, 1, num_params_used, Nested_Like, Nested_Prior, &
initial_cosmo, replace_ellipsoid, cs)
                                                                                                   
  CALL print_pset(cs, trim(rootname)//"_nested_pset.dat")
                                                                                                   
   ev = evidence(cs)
   PRINT *,"Evidence",ev(1),"+/-",ev(2)
   ! Print sampling efficiency.
   CALL print_sampling_eff()
                                                                                                   
end subroutine Nested_Evidence

! Wrapper around Likelihood Function
real(rk) function Nested_Like(data, par)                                                                                       
  ! The array of data (dummy ATM)
  REAL(rk), DIMENSION(:,:) :: data
 ! Parameters
  REAL(rk), DIMENSION(:) :: par
REAL(rk) :: net_params(1:2*num_params_used)
logical within_range                                                                                                   
integer i

  ! Add P vector into ParamSet Type to use in GetLogLike
  IF(size(par) /= num_params_used) THEN
     PRINT *,"Eeek param vector length wrong"
  ELSE
!!!!!!!!!!!!WITHIN_RANGE STUFF!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!6param 4sigma region!!!!!!!!!!!!!!!!!!!!!!!!!!
!                net_params(1)=0.0195495
!                net_params(2)=0.0959060
!                net_params(3)=1.02596
!                net_params(4)=0.0100000
!                net_params(5)=0.890362
!                net_params(6)=2.78152
!                                                                                          
!                net_params(7)=0.0243130
!                net_params(8)=0.122320
!                net_params(9)=1.04990
!                net_params(10)=0.165190
!                net_params(11)=0.988530
!                net_params(12)=3.17939
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!10param 4 sigma region!!!!!!!!!!!!
net_params(1) = 0.0205637d0
net_params(2) = 0.0801418d0
net_params(3) = 1.03216d0
net_params(4) = 0.0100661d0
net_params(5) = -0.145745d0
net_params(6) = 0.000000d0
net_params(7) = -1.50000d0
net_params(8) = 0.90911d0
net_params(9) = 2.81763d0
net_params(10) = 0.000000d0
net_params(11)=0.0262164d0
net_params(12)=0.141981d0
net_params(13)=1.05557d0
net_params(14)=0.181053d0
net_params(15)= 0.0719305d0
net_params(16) = 0.136317d0
net_params(17) = -0.0955112d0
net_params(18)=1.07102d0
net_params(19)=3.22724d0
net_params(20) = 0.606624d0



                FScales%PMin = Scales%PMin
                FScales%PMax = Scales%PMax
                !do i=1, num_params_used
                !        FScales%PMin(params_used(i)) = net_params(i)
                !        FScales%PMax(params_used(i)) = net_params(i+num_params_used)
                         
		!end do
     Trial % P(params_used) = par
if (any(Trial%P > FScales%PMax) .or. any(Trial%P < FScales%PMin)) then
        within_range = .false.
        write(*,*) "Not using Cosmonet"
else
        within_range = .true.
end if
                                                                                          
if (within_range) then
        Nested_Like  = -1.0*COSMONET_GetLogLike(Trial, time, first_time)
else
        Nested_Like = -1.0*GetLogLike(Trial)
end if


     !Nested_Like = -1.0 * GetLogLike(Trial)
  END IF
                                                                                                   
                                                                                                   
end function Nested_Like
                                                                                                   
! Prior function
real(rk) function Nested_Prior(par)
  ! Parameters
  REAL(rk), DIMENSION(:) :: par
                                                                                                   
  INTEGER :: i
  REAL(rk) :: vol
  vol = 1.0
                                                                                                   
  !PRINT *,"Pp",par
                                                                                                   
  DO i=1,num_params_used
    vol = vol * (p_max(params_used(i)) - p_min(params_used(i)))
    IF(par(i) < p_min(params_used(i)) .OR. par(i) > p_max(params_used(i))) THEN
   !   PRINT *,"Rejected"
      Nested_Prior = min_inf
      RETURN
    END IF
  END DO
                                                                                                   
                                                                                                   
  Nested_Prior = -1.0 * log(vol)
                                                                                                   
 !PRINT *,"Pp1",Nested_Prior
                                                                                                   
end function Nested_Prior
                                                                                                   
                                                                                                   
                                                                                                   
! Tophat prior method for initial selection is cosmology.
SUBROUTINE initial_cosmo(tree, n, data, ddata, dpars, L, X)
                                                                                                   
  ! The point tree to add to.
  TYPE(pnode), POINTER :: tree
  ! Number of points to add.
  INTEGER, INTENT(IN) :: n
  ! Array of data to pass to likelihood function.
  REAL(rk), INTENT(IN), DIMENSION(:,:) :: data
  ! Dimensionality of data space.
  INTEGER, INTENT(IN) :: ddata
  ! Dimensionality of parameter space.
  INTEGER, INTENT(IN) :: dpars
                                                                                                   
  INTERFACE
    ! The Likelihood function
     REAL(rk) FUNCTION L(pdata, ppars1)
       REAL(rk), DIMENSION(:,:) :: pdata
       REAL(rk), DIMENSION(:) :: ppars1

     END FUNCTION L
                                                                                                   
     ! The Prior function
     REAL(rk) FUNCTION X(ppars)
       REAL(rk), DIMENSION(:) :: ppars
     END FUNCTION X
                                                                                                   
  END INTERFACE
                                                                                                   
                                                                                                   
  INTEGER :: i
  REAL(rk), DIMENSION(:), ALLOCATABLE :: point
  REAL(rk), DIMENSION(:), ALLOCATABLE :: widths
  REAL(rk) :: like
  ALLOCATE(widths(num_params_used))
                                                                                                   
  ALLOCATE(point(dpars))
  widths = p_max(params_used) - p_min(params_used)
  !PRINT *,"Widths",widths
  DO i=1,n
                                                                                                   
     CALL rng_uniform_arr(point, dpars)
    ! PRINT *,"P1",point
     point = (widths*point) + p_min(params_used)
     like = L(data, point)
                                                                                                   
     DO WHILE(isnan(like))
        CALL rng_uniform_arr(point, dpars)
        ! PRINT *,"P1",point
        point = (widths*point) + p_min(params_used)
        like = L(data, point)
     END DO
     WRITE(*,"(A,20G20.10)") "Point",point,like
     CALL insert_node(tree, X(point),like, point)
                                                                                                   
  END DO
END SUBROUTINE initial_cosmo
                                                                                                   
                                                                                                   
                                                                                                   
! Wrapper around Prior Function
real(rk) function Nested_Prior_Old(par)
                                                                                                   
  ! Parameters
  REAL(rk), DIMENSION(:) :: par
                                                                                                   
  TYPE(CMBParams) :: CMB
                                                                                                   
  ! Generate a CMBParams Type for use in GetLogPrior
  CALL ParamsToCMBParams(par, CMB)
                                                                                                   
  Nested_Prior_Old = -1.0 * GetLogPrior(CMB, Trial % Info)
                                                                                                   
end function Nested_Prior_Old

 
end module MonteCarlo
